/*
 * Copyright (C) 2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "gcc-compat.h"
#include "para.h"

#include "filter.cmdline.h"
#include "list.h"
#include "filter.h"

#define INBUF_SIZE 16 * 1024

static struct filter_chain_info filter_chain_info_struct;
static struct filter_chain_info *fci = &filter_chain_info_struct;

struct gengetopt_args_info conf;

__printf(2,3) void para_log(int ll, char* fmt,...)
{
	va_list argp;

	/* ignore log message if loglevel is not high enough */
	if (ll < conf.loglevel_arg)
		return;
	va_start(argp, fmt);
	vfprintf(stderr, fmt, argp);
	va_end(argp);
}

static char *inbuf;
static size_t loaded;

extern struct filter filters[];

static int init_active_filter_list(void)
{
	struct filter_node *fn;
	int i;

	INIT_LIST_HEAD(&fci->filters);

	fci->inbuf = inbuf;
	fci->in_loaded = &loaded;

	for (i = 0; i < conf.filter_given; i++) {
		char *fa = para_strdup(conf.filter_arg[i]);
		fn = check_filter_arg(fa);
		if (!fn) {
			free(fa);
			return -1;
		}
		PARA_DEBUG_LOG("adding %s to filter chain\n", fn->filter->name);
		list_add_tail(&fn->node, &fci->filters);
	}
	if (!list_empty(&fci->filters))
		return 1;
	PARA_ERROR_LOG("%s", "no filters given, aborting\n");
	return -1;
}

static int open_filters(void)
{
	struct filter_node *fn;

	list_for_each_entry(fn, &fci->filters, node) {
		int ret = fn->filter->open(fn);
		PARA_INFO_LOG("opened %s filter\n", fn->filter->name);
		if (ret < 0)
			return -1;
		fci->outbuf = fn->buf;
		fci->out_loaded = &fn->loaded;
	}
	return 1;
}

static int parse_config(int argc, char *argv[])
{
	static char *cf; /* config file */
        struct stat statbuf;
	int i;

	if (cmdline_parser(argc, argv, &conf))
		return -1;
	if (!cf) {
		char *home = para_homedir();
		cf = make_message("%s/.paraslash/filter.conf", home);
		free(home);
	}
        if (!stat(cf, &statbuf)) {
		if (cmdline_parser_configfile(cf, &conf, 0, 0, 0)) {
			PARA_ERROR_LOG("%s", "parse error in config file\n");
			return -2;
		}
	}
	if (!conf.list_filters_given)
		return 1;
	printf("available filters: ");
	for (i = 0; filters[i].name; i++)
		printf("%s%s", i? " " : "", filters[i].name);
	printf("\nTry para_filter -f<filtername>:-h for help on <filtername>\n");
	exit(EXIT_SUCCESS);
}

int main(int argc, char *argv[])
{
	int converted, ret, eof = 0;
	char *ib, *ob; /* input/output buffer */
	size_t *il, *ol; /* number of loaded bytes in input/output buffer */

	filter_init(filters);
	if (parse_config(argc, argv) < 0)
		exit(EXIT_FAILURE);
	inbuf = para_malloc(INBUF_SIZE);
	if (init_active_filter_list() < 0)
		exit(EXIT_FAILURE);
	open_filters();
	ib = fci->inbuf;
	ob = fci->outbuf;
	il = fci->in_loaded;
	ol = fci->out_loaded;
again:
	if (*il < INBUF_SIZE / 2 && !eof) {
		ret  = read(STDIN_FILENO, ib + *il, INBUF_SIZE - *il);
//		PARA_DEBUG_LOG("read %d/%d\n", ret, INBUF_SIZE - *il);
		if (ret < 0)
			goto out;
		if (!ret)
			eof = 1;
		*il += ret;
	}
	converted = filter_io(fci);
	ret = -1;
	if (converted < 0)
		goto out;
//	PARA_DEBUG_LOG("loaded: %d in, %d out\n", *il, *ol);
	if (*ol) {
		ret = write(STDOUT_FILENO, ob, *ol);
//		PARA_DEBUG_LOG("wrote %d/%d\n", ret, *ol);
		if (ret <= 0)
			goto out;
		*ol -= ret;
		if (*ol) {
			PARA_NOTICE_LOG("short write: %d bytes left\n", *ol);
			memmove(ob, ob + ret, *ol);
		}
	}
	if (!eof || converted)
		goto again;
	ret = 0;
out:
	close_filters(fci);
	return ret;
}
