/*
 * Copyright (C) 2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/* This is a very simple oggvorbis decoder which decodes stdin to stdout.
 */

#include "gcc-compat.h"
#include "para.h"

#include "oggdec_filter.cmdline.h"
#include "list.h"
#include "filter.h"

#include <vorbis/vorbisfile.h>
#define BITS 16
#define ENDIAN 0
#define SIGN 1

struct private_ogg_data {
	size_t prebuffer_size;
	OggVorbis_File *vf;
	char *inbuf;
	size_t inbuf_len, converted;
	size_t loaded; /* how many bytes are really loaded */
	struct gengetopt_args_info conf;
};

static size_t cb_read(void *buf, size_t size, size_t nmemb, void *datasource)
{
	struct private_ogg_data *pod = datasource;
	size_t ret, have = pod->inbuf_len - pod->converted;
	char *p = pod->inbuf + pod->converted;

//	PARA_DEBUG_LOG("pod = %p\n", pod);
	PARA_INFO_LOG("vorbis requests %d bytes, have %d\n", size * nmemb, have);
	if (pod->inbuf_len < size) {
		errno = EAGAIN;
		return -1;
	}
	ret = MIN(nmemb, have / size);
	memcpy(buf, p, ret * size);
	pod->converted += ret * size;
	return ret * size;
}

/**
 * cb_seek -- custom data seeking function
 * Since we want the data source to be treated as unseekable at all
 * times, the provided seek callback always returns -1 (failure).
 **/
static int cb_seek(void *datasource, ogg_int64_t offset, int whence)
{
	return -1;
}

static int cb_close(void *datasource)
{
	return 0;
}

static const ov_callbacks ovc = {
	.read_func = cb_read,
	.seek_func = cb_seek,
	.close_func = cb_close,
	/*
	 * The tell function need not be provided if 
	 * the data IO abstraction is not seekable
	 */
	.tell_func = NULL
};

static int ogg_open(struct filter_node *fn)
{
	struct private_ogg_data *pod = para_calloc(
		sizeof(struct private_ogg_data));

	oggdec_cmdline_parser(fn->argc, fn->argv, &pod->conf);
	fn->private_data = pod;
	fn->bufsize = pod->conf.bufsize_arg;
	fn->buf = para_malloc(fn->bufsize);
	return 1;
}

static void ogg_close(struct filter_node *fn)
{
	struct private_ogg_data *pod = fn->private_data;
	if (pod->vf) {
		ov_clear(pod->vf);
		free(pod->vf);
		pod->vf = NULL;
	}
	free(fn->buf);
	fn->buf = NULL;
	free(fn->private_data);
	fn->private_data = NULL;
}

static ssize_t ogg_convert(char *inbuffer, size_t len, struct filter_node *fn,
		struct filter_chain_info *fci)
{
	ssize_t ret;
	struct private_ogg_data *pod = fn->private_data;
	int new_buffer = 0;

	if (!pod->prebuffer_size)
		pod->loaded = fn->loaded;
	if (pod->inbuf_len == pod->converted) {
		PARA_INFO_LOG("new input buffer: %d bytes\n", len);
		pod->inbuf_len = len;
		pod->inbuf = inbuffer;
		pod->converted = 0;
		new_buffer = 1;
	}
	if (!pod->vf) {
		pod->vf = para_malloc(sizeof(struct OggVorbis_File));
		PARA_NOTICE_LOG("%s", "opening ov callbacks\n");
		ret = ov_open_callbacks(pod, pod->vf,
			NULL, /* no initial buffer */
			0, /* no initial bytes */ 
			ovc); /* the ov_open_callbacks */
		if (ret < 0)
			return -1;
		fci->channels = ov_info(pod->vf, 0)->channels;
		fci->samplerate = ov_info(pod->vf, 0)->rate;
		PARA_NOTICE_LOG("%d channels, %d Hz\n", fci->channels, fci->samplerate);
	}
	ret = ov_read(pod->vf, fn->buf + pod->loaded, fn->bufsize - pod->loaded,
		ENDIAN, BITS / 8, SIGN, NULL);
	PARA_DEBUG_LOG("ov_read: %d/%d, converted: %d/%d\n", ret,
			fn->bufsize - pod->loaded, pod->converted, pod->inbuf_len);
	if (ret == OV_HOLE && !pod->loaded) {
		pod->prebuffer_size = pod->conf.bufsize_arg;
		PARA_NOTICE_LOG("detected hole in data, will prebuffer up to %d bytes\n", pod->prebuffer_size);
		ret = 0;
	}
	if (ret < 0)
		return -1;
	pod->loaded += ret;
	if (pod->prebuffer_size) {
		ogg_int64_t raw = ov_raw_tell(pod->vf);
		if (pod->loaded >= pod->prebuffer_size || raw > 25000) {
			PARA_INFO_LOG("end of prebuffering (have %d bytes)\n", pod->loaded);
			fn->loaded = pod->loaded;
			pod->prebuffer_size = 0;
		}
	} else
		fn->loaded += ret;
	if (pod->inbuf_len == pod->converted) {
		ret = pod->inbuf_len;
		PARA_DEBUG_LOG("returning %d\n", ret);
	} else
		ret = 0;
	return ret;
}

static int oggdec_parse_config(struct filter_node *fn)
{
	struct gengetopt_args_info tmp;
	return oggdec_cmdline_parser(fn->argc, fn->argv, &tmp)? -1 : 1;
}

int oggdec_init(struct filter *f)
{
	f->open = ogg_open;
	f->close = ogg_close;
	f->convert = ogg_convert;
	f->print_help = oggdec_cmdline_parser_print_help;
	f->parse_config = oggdec_parse_config;
	return 1;
}
